@@ -14,6 +14,7 @@
 ;;; The CLtL way
 
 (defpackage :xlib (:use :lisp #+(and kcl clos-conditions) :conditions))
+(in-package :xlib)
 
 #+(and (or kcl ibcl) (not clx-ansi-common-lisp))
 (shadow
@@ -21,9 +22,6 @@
     rational
     ))
 
-#+clisp
-(shadow :boolean)
-
 #+(and lispm (not clx-ansi-common-lisp))
 (import
   '(
