@@ -1,11 +1,24 @@
 /*
  * system calls
- * Copyright (C) 2003 Sam Steingold
+ * Copyright (C) 2003-2004 Sam Steingold
  * GPL2
  */
 
 #include "config.h"
 
+#if defined(_WIN32)
+/* need this for CreateHardLink to work */
+# define WINVER 0x0500
+/* get ASCII functions */
+# undef UNICODE
+#endif
+#if defined(__CYGWIN__)
+# define UNIX_CYGWIN32
+# undef UNICODE
+#endif
+
+#include "clisp.h"
+
 #if defined(TIME_WITH_SYS_TIME)
 # include <sys/time.h>
 # include <time.h>
@@ -28,23 +41,17 @@
 #if defined(HAVE_SYS_STAT_H)
 # include <sys/stat.h>
 #endif
-
-#if defined(_WIN32)
-/* need this for CreateHardLink to work */
-# define WINVER 0x0500
-/* get ASCII functions */
-# undef UNICODE
+#if defined(HAVE_SYS_TYPES_H)
+# include <sys/types.h>
 #endif
-#if defined(__CYGWIN__)
-# define UNIX_CYGWIN32
-# undef UNICODE
+#if defined(HAVE_SYS_STATVFS_H)
+# include <sys/statvfs.h>
 #endif
 
-#include "clisp.h"
+#include <stdio.h>             /* for BUFSIZ */
 
 /* #define DEBUG */
 #if defined(DEBUG)
-# include <stdio.h>
 extern object nobject_out (FILE* stream, object obj);
 # define XOUT(obj,label)                                                \
   (printf("[%s:%d] %s: %s:\n",__FILE__,__LINE__,STRING(obj),label),     \
@@ -55,7 +62,7 @@ extern object nobject_out (FILE* stream, object obj);
 # define XOUT(o,l)
 #endif
 
-DEFMODULE(syscalls,"POSIX");
+DEFMODULE(syscalls,"POSIX")
 
 #if defined(HAVE_FCNTL) || defined(WIN32_NATIVE)
 /* we use posix fcntl() on unix and win32 LockFileEx() on win32.
@@ -92,7 +99,11 @@ DEFUN(POSIX::STREAM-LOCK, stream lockp &key BLOCK SHARED START LENGTH)
 #    if defined(WIN32_NATIVE)
       LARGE_INTEGER size;
       begin_system_call();
-      failed_p = (!GetFileSizeEx(fd,&size));
+      size.LowPart = GetFileSize(fd,&size.HighPart);
+      /* Value returned can be (LONG) -1 even on success,
+         check the last error code */
+      if ((LONG)size.LowPart == -1) failed_p = GetLastError();
+        else failed_p = 0;
       end_system_call();
       if (failed_p) goto error;
       length = size.LowPart;
@@ -278,6 +289,122 @@ DEFUN(POSIX::SYSCONF,)
 }
 #endif /* HAVE_SYSCONF */
 
+#if defined(HAVE_CONFSTR)
+DEFUN(POSIX::CONFSTR,)
+{ /* Lisp interface to confstr(3c) */
+  size_t res;
+  char buf[BUFSIZ];
+
+#define CS_S(cmd) \
+  begin_system_call(); res = confstr(cmd,buf,BUFSIZ); end_system_call(); \
+  if (res == 0) pushSTACK(T);                                           \
+  else if (res <= BUFSIZ) pushSTACK(asciz_to_string(buf,GLO(misc_encoding))); \
+  else {                                                                \
+    begin_system_call();                                                \
+    { char *tmp = alloca(res);                                          \
+      confstr(cmd,tmp,res);                                             \
+      end_system_call();                                                \
+      pushSTACK(asciz_to_string(tmp,GLO(misc_encoding)));               \
+    }}
+
+#if defined(_CS_PATH)
+  CS_S(_CS_PATH);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_ILP32_OFF32_CFLAGS)
+  CS_S(_CS_POSIX_V6_ILP32_OFF32_CFLAGS);
+#elif defined(_CS_XBS5_ILP32_OFF32_CFLAGS)
+  CS_S(_CS_XBS5_ILP32_OFF32_CFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_ILP32_OFF32_LDFLAGS)
+  CS_S(_CS_POSIX_V6_ILP32_OFF32_LDFLAGS);
+#elif defined(_CS_XBS5_ILP32_OFF32_LDFLAGS)
+  CS_S(_CS_XBS5_ILP32_OFF32_LDFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_ILP32_OFF32_LIBS)
+  CS_S(_CS_POSIX_V6_ILP32_OFF32_LIBS);
+#elif defined(_CS_XBS5_ILP32_OFF32_LIBS)
+  CS_S(_CS_XBS5_ILP32_OFF32_LIBS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_ILP32_OFFBIG_CFLAGS)
+  CS_S(_CS_POSIX_V6_ILP32_OFFBIG_CFLAGS);
+#elif defined(_CS_XBS5_ILP32_OFFBIG_CFLAGS)
+  CS_S(_CS_XBS5_ILP32_OFFBIG_CFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_ILP32_OFFBIG_LDFLAGS)
+  CS_S(_CS_POSIX_V6_ILP32_OFFBIG_LDFLAGS);
+#elif defined(_CS_XBS5_ILP32_OFFBIG_LDFLAGS)
+  CS_S(_CS_XBS5_ILP32_OFFBIG_LDFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_ILP32_OFFBIG_LIBS)
+  CS_S(_CS_POSIX_V6_ILP32_OFFBIG_LIBS);
+#elif defined(_CS_XBS5_ILP32_OFFBIG_LIBS)
+  CS_S(_CS_XBS5_ILP32_OFFBIG_LIBS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_LP64_OFF64_CFLAGS)
+  CS_S(_CS_POSIX_V6_LP64_OFF64_CFLAGS);
+#elif defined(_CS_XBS5_LP64_OFF64_CFLAGS)
+  CS_S(_CS_XBS5_LP64_OFF64_CFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_LP64_OFF64_LDFLAGS)
+  CS_S(_CS_POSIX_V6_LP64_OFF64_LDFLAGS);
+#elif defined(_CS_XBS5_LP64_OFF64_LDFLAGS)
+  CS_S(_CS_XBS5_LP64_OFF64_LDFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_LP64_OFF64_LIBS)
+  CS_S(_CS_POSIX_V6_LP64_OFF64_LIBS);
+#elif defined(_CS_XBS5_LP64_OFF64_LIBS)
+  CS_S(_CS_XBS5_LP64_OFF64_LIBS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_LPBIG_OFFBIG_CFLAGS)
+  CS_S(_CS_POSIX_V6_LPBIG_OFFBIG_CFLAGS);
+#elif defined(_CS_XBS5_LPBIG_OFFBIG_CFLAGS)
+  CS_S(_CS_XBS5_LPBIG_OFFBIG_CFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_LPBIG_OFFBIG_LDFLAGS)
+  CS_S(_CS_POSIX_V6_LPBIG_OFFBIG_LDFLAGS);
+#elif defined(_CS_XBS5_LPBIG_OFFBIG_LDFLAGS)
+  CS_S(_CS_XBS5_LPBIG_OFFBIG_LDFLAGS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_LPBIG_OFFBIG_LIBS)
+  CS_S(_CS_POSIX_V6_LPBIG_OFFBIG_LIBS);
+#elif defined(_CS_XBS5_LPBIG_OFFBIG_LIBS)
+  CS_S(_CS_XBS5_LPBIG_OFFBIG_LIBS);
+#else
+  pushSTACK(NIL);
+#endif
+#if defined(_CS_POSIX_V6_WIDTH_RESTRICTED_ENVS)
+  CS_S(_CS_POSIX_V6_WIDTH_RESTRICTED_ENVS);
+#else
+  pushSTACK(NIL);
+#endif
+  funcall(`POSIX::MAKE-CONFSTR`,14);
+}
+#endif /* HAVE_CONFSTR */
+
 #if defined(HAVE_SYS_RESOURCE_H)
 # include <sys/resource.h>
 #endif
@@ -547,10 +674,10 @@ DEFUN(POSIX::USER-DATA, user)
 
 DEFUN(POSIX::FILE-STAT, file &optional linkp)
 { /* Lisp interface to stat(2), lstat(2) and fstat(2)
- the first arg can be: file stream, pathname, string, symbol, number.
+ the first arg can be a pathname designator or a file descriptor designator
  the return value is the FILE-STAT structure */
-  bool link_p = missingp(STACK_0); skipSTACK(1);
-  object file = popSTACK();
+  bool link_p = missingp(STACK_0);
+  object file = (skipSTACK(1), popSTACK());
   struct stat buf;
 
   if (builtin_stream_p(file)) {
@@ -592,10 +719,200 @@ DEFUN(POSIX::FILE-STAT, file &optional linkp)
   funcall(`POSIX::MAKE-FILE-STAT`,14);
 }
 #endif  /* fstat lstat fstat */
+
+#if defined(HAVE_FSTATVFS) && defined(HAVE_STATVFS)
+/* there is also a legacy interface (f)statfs()
+   which is not POSIX and is not supported */
+
+DEFUN(POSIX::STAT-VFS, file)
+{ /* Lisp interface to statvfs(2), fstatvfs(2)
+ the first arg can be a pathname designator or a file descriptor designator
+ the return value is the STAT-VFS structure */
+  object file = popSTACK();
+  struct statvfs buf;
+
+  if (builtin_stream_p(file)) {
+    pushSTACK(file);
+    funcall(L(built_in_stream_open_p),1);
+    if (!nullp(value1)) { /* open stream ==> use FD */
+      begin_system_call();
+      if (fstatvfs(stream_lend_handle(file,true,NULL),&buf) < 0) OS_error();
+      end_system_call();
+    } else goto stat_pathname;
+  } else if (integerp(file)) {
+    begin_system_call();
+    if (fstatvfs(I_to_L(file),&buf) < 0) OS_error();
+    end_system_call();
+  } else { stat_pathname:
+    file = whole_namestring(file);
+    with_string_0(file,GLO(pathname_encoding),namez, {
+      begin_system_call();
+      if (statvfs(namez,&buf) < 0) OS_error();
+      end_system_call();
+    });
+  }
+
+  pushSTACK(file);                  /* the object statvfs'ed */
+  /* FIXME: Should use ulong_to_I for most of these, and map -1 to NIL. */
+  pushSTACK(UL_to_I(buf.f_bsize));  /* file system block size */
+  pushSTACK(UL_to_I(buf.f_frsize)); /* fundamental file system block size */
+  pushSTACK(UL_to_I(buf.f_blocks)); /* total # of blocks on file system */
+  pushSTACK(UL_to_I(buf.f_bfree));  /* total number of free blocks */
+  pushSTACK(UL_to_I(buf.f_bavail)); /* # of free blocks available to
+                                       non-privileged processes */
+  pushSTACK(UL_to_I(buf.f_files));  /* total # of file serial numbers */
+  pushSTACK(UL_to_I(buf.f_ffree));  /* total # of free file serial numbers */
+  pushSTACK(UL_to_I(buf.f_favail)); /* # of file serial numbers available to
+                                       non-privileged processes */
+#if HAVE_SCALAR_FSID
+  pushSTACK(UL_to_I(buf.f_fsid));   /* file system ID */
+#else
+  /* On Linux, f_fsid of 'struct statfs' is a struct consisting of two ints.
+     With glibc <= 2.1, f_fsid of 'struct statvfs' is the same. We are
+     prepared to return one number only, so we just return the first int.
+     This matches the behaviour of glibc >= 2.2 on 32-bit platforms. */
+  pushSTACK(UL_to_I(*(uintL*)&buf.f_fsid));   /* file system ID */
+#endif
+  { /* bit mask of f_flag values */
+    unsigned long count = 0;
+#  ifdef ST_RDONLY
+    if (buf.f_flag & ST_RDONLY) { pushSTACK(S(Kread_only)); count++; }
+#  endif
+#  ifdef ST_NOSUID
+    if (buf.f_flag & ST_NOSUID) { pushSTACK(`:NO-SUID`); count++; }
+#  endif
+#  ifdef ST_NOTRUNC
+    if (buf.f_flag & ST_NOTRUNC) { pushSTACK(`:NO-TRUNCATE`); count++; }
+#  endif
+    if (count) {
+      object res = listof(count);
+      pushSTACK(res);
+    } else pushSTACK(NIL);
+  }
+  pushSTACK(UL_to_I(buf.f_namemax));/* maximum filename length */
+  funcall(`POSIX::MAKE-STAT-VFS`,12);
+}
+
+#endif  /* fstatvfs statvfs */
+
 #endif /* UNIX */
 
 /* COPY-FILE related functions. */
 
+#if defined(WIN32_NATIVE)
+
+/* Pointers to functions unavailable on windows 95, 98, ME */
+
+typedef BOOL (WINAPI * CreateHardLinkFuncType) ( LPCTSTR lpFileName, LPCTSTR lpExistingFileName,
+   LPSECURITY_ATTRIBUTES lpSecurityAttributes);
+static CreateHardLinkFuncType CreateHardLinkFunc = NULL;
+
+typedef BOOL (WINAPI * BackupWriteFuncType) (HANDLE hFile, LPBYTE lpBuffer, DWORD nNumberOfBytesToWrite,
+   LPDWORD lpNumberOfBytesWritten, BOOL bAbort, BOOL bProcessSecurity, LPVOID *lpContext);
+static BackupWriteFuncType BackupWriteFunc = NULL;
+
+static HMODULE kernel32 = NULL;
+
+/* Checks if it's safe to call OldHardLink */
+static BOOL OldHardLinkGuard () {
+  OSVERSIONINFO vi;
+  if (BackupWriteFunc == NULL) return FALSE;
+  vi.dwOSVersionInfoSize = sizeof(vi);
+  if (!GetVersionEx(&vi)) return FALSE;
+  return vi.dwPlatformId == VER_PLATFORM_WIN32_NT;
+}
+
+/* From knowledge base article Q234727
+   This approach works on NT >= 3.51. */
+static BOOL OldHardLink( LPCTSTR source, LPCTSTR dest ) {
+
+   WCHAR  wsource[ MAX_PATH + 1 ];
+   WCHAR  wdest[ MAX_PATH + 1 ];
+   WCHAR  wdestfull[ MAX_PATH + 1 ];
+   LPWSTR wdestfullfile;
+
+   HANDLE hFileSource;
+
+   WIN32_STREAM_ID StreamId;
+   DWORD dwBytesWritten;
+   LPVOID lpContext;
+   DWORD cbPathLen;
+   DWORD StreamHeaderSize;
+
+   BOOL bSuccess;
+
+   /* convert from ANSI to UNICODE */
+   if (MultiByteToWideChar(CP_ACP, MB_ERR_INVALID_CHARS /*error on invalid chars*/,
+     dest, -1/*null terminated*/, wdest, MAX_PATH + 1) == 0) return FALSE;
+
+   /* open existing file that we link to */
+   hFileSource = CreateFile(source, FILE_WRITE_ATTRIBUTES,
+     FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
+     NULL, /* sa */ OPEN_EXISTING, 0, NULL );
+
+   if (hFileSource == INVALID_HANDLE_VALUE) return FALSE;
+
+   /* validate and sanitize supplied link path and use the result
+      the full path MUST be Unicode for BackupWrite */
+   cbPathLen = GetFullPathNameW( wdest , MAX_PATH, wdestfull, &wdestfullfile);
+
+   if (cbPathLen == 0) return FALSE;
+
+   cbPathLen = (cbPathLen + 1) * sizeof(WCHAR); // adjust for byte count
+
+   /* prepare and write the WIN32_STREAM_ID out */
+   lpContext = NULL;
+
+   StreamId.dwStreamId = BACKUP_LINK;
+   StreamId.dwStreamAttributes = 0;
+   StreamId.dwStreamNameSize = 0;
+   StreamId.Size.HighPart = 0;
+   StreamId.Size.LowPart = cbPathLen;
+
+   /* compute length of variable size WIN32_STREAM_ID */
+   StreamHeaderSize = (LPBYTE)&StreamId.cStreamName - (LPBYTE)&StreamId
+                      + StreamId.dwStreamNameSize ;
+
+   bSuccess = BackupWriteFunc(hFileSource,
+                         (LPBYTE)&StreamId,  /* buffer to write */
+                         StreamHeaderSize,   /* number of bytes to write */
+                         &dwBytesWritten,
+                         FALSE,              /* don't abort yet */
+                         FALSE,              /* don't process security */
+                         &lpContext);
+   bSuccess &= BackupWriteFunc(hFileSource,(LPBYTE)wdestfull, cbPathLen,
+        &dwBytesWritten, FALSE, FALSE, &lpContext);
+   /* free context */
+   bSuccess &= BackupWriteFunc(hFileSource,NULL,0,&dwBytesWritten,TRUE, FALSE,
+        &lpContext);
+   CloseHandle( hFileSource );
+   return bSuccess;
+}
+#endif
+
+void module__syscalls__init_function_2 (module_t* module) {
+#if defined(WIN32_NATIVE)
+  kernel32 = LoadLibrary ("kernel32.dll");
+  if (kernel32 != NULL) {
+    CreateHardLinkFunc = (CreateHardLinkFuncType)
+      GetProcAddress (kernel32, "CreateHardLinkA");
+    BackupWriteFunc = (BackupWriteFuncType)
+      GetProcAddress (kernel32, "BackupWrite");
+  }
+#endif
+}
+
+#if defined(WIN32_NATIVE)
+static inline int MkHardLink (char* old_pathstring, char* new_pathstring) {
+  if (CreateHardLinkFunc != NULL)
+    return CreateHardLinkFunc(new_pathstring,old_pathstring,NULL);
+  if (OldHardLinkGuard())
+    return OldHardLink(old_pathstring,new_pathstring);
+  SetLastError(ERROR_INVALID_FUNCTION); /* or what ? */
+  return 0;
+}
+#endif
+
 /* Hard/Soft Link a file
  > old_pathstring: old file name, ASCIZ-String
  > new_pathstring: new file name, ASCIZ-String
@@ -608,7 +925,7 @@ DEFUN(POSIX::FILE-STAT, file &optional linkp)
 static inline void hardlink_file (char* old_pathstring, char* new_pathstring) {
   begin_system_call();
 # if defined(WIN32_NATIVE)
-  if (CreateHardLink(new_pathstring,old_pathstring,NULL) == 0)
+  if (MkHardLink(old_pathstring,new_pathstring) == FALSE)
     if (GetLastError() == ERROR_FILE_NOT_FOUND)
 # else
   if (link(old_pathstring,new_pathstring) < 0)
@@ -685,7 +1002,7 @@ static void copy_attributes_and_close () {
   goto close_success;
 #  endif  /* fstat() */
 # else
-  /*** FIXME: windows? amiga? riscos? ***/
+  /*** FIXME: windows? ***/
 # endif
  close_success:
   builtin_stream_close(&STACK_0);
@@ -747,9 +1064,9 @@ static void copy_file_low (object source, object dest,
     Handle fd_in = stream_lend_handle(STACK_1,true,NULL);
     Handle fd_ou = stream_lend_handle(STACK_0,false,NULL);
     while ((bytes_read = read_helper(fd_in,buffer,strm_buffered_bufflen,
-                                     true))) {
+                                     false))) {
       total_count += bytes_read;
-      full_write(fd_ou,buffer,bytes_read);
+      write_helper(fd_ou,buffer,bytes_read,false);
     }
   }
   if (!preserve_p) {
@@ -801,7 +1118,7 @@ static inline object copy_method_object (copy_method_t method) {
   }
 }
 
-/* copy just one file: source --> dest (both strings)
+/* copy just one file: source --> dest (both STRINGs, NIL or PATHNAME)
    can trigger GC */
 static void copy_one_file (object source, object src_path,
                            object dest, object dest_path,
@@ -826,20 +1143,17 @@ static void copy_one_file (object source, object src_path,
   }
 
   pushSTACK(STACK_0); funcall(L(probe_file),1);
-  if (!nullp(value1)) { /* destination exists */
-    pushSTACK(value1); funcall(L(truename),1);
-    pushSTACK(value1); dest = value1; STACK_2 = dest;
+  if (!nullp(value1)) { /* destination exists; value1 == truename */
+    pushSTACK(value1); STACK_2 = dest = value1;
     /* STACK: 0=dest_true; 1=dest_path; 2=dest; 3=src_path; 4=src */
     switch (if_exists) {
       case IF_EXISTS_NIL: skipSTACK(5); return;
       case IF_EXISTS_APPEND:
-        if (method != COPY_METHOD_COPY) {
-          pushSTACK(`:APPEND`);
-          pushSTACK(copy_method_object(method));
-          pushSTACK(`POSIX::COPY-FILE`);
-          fehler(error,GETTEXT("~: ~ forbids ~"));
-        }
-        break;
+        /* we know that method != COPY_METHOD_COPY - handled above! */
+        pushSTACK(`:APPEND`);
+        pushSTACK(copy_method_object(method));
+        pushSTACK(`POSIX::COPY-FILE`);
+        fehler(error,GETTEXT("~: ~ forbids ~"));
       case IF_EXISTS_OVERWRITE:
       case IF_EXISTS_SUPERSEDE:
       case IF_EXISTS_RENAME_AND_DELETE:
@@ -856,7 +1170,7 @@ static void copy_one_file (object source, object src_path,
         break;
       default: NOTREACHED;
     }
-  } else pushSTACK(NIL); /* destination does not exist */
+  } else pushSTACK(STACK_0); /* destination does not exist, use dest_path */
 
   pushSTACK(STACK_3); funcall(L(probe_file),1);
   if (nullp(value1)) { /* source does not exist */
@@ -877,7 +1191,6 @@ static void copy_one_file (object source, object src_path,
   }
 
   /* stack layout: 0=src_true; 1=dest_true ... */
-  dest = STACK_3; /* restore true namestring */
   switch (method) {
     case COPY_METHOD_RENAME:
       pushSTACK(STACK_0); pushSTACK(STACK_2); funcall(L(rename_file),2);
@@ -885,6 +1198,7 @@ static void copy_one_file (object source, object src_path,
       break;
     case COPY_METHOD_SYMLINK:
 #    if defined(HAVE_SYMLINK)
+      dest = whole_namestring(STACK_1);
       /* use the original argument, not the truename here,
          so that the user can create relative symlinks */
       source = stringp(STACK_5) ? (object)STACK_5 : whole_namestring(STACK_4);
@@ -897,6 +1211,8 @@ static void copy_one_file (object source, object src_path,
       /* FALLTHROUGH if no symlinks */
     case COPY_METHOD_HARDLINK:
 #    if defined(HAVE_LINK)
+      dest = whole_namestring(STACK_1);
+      source = whole_namestring(STACK_0);
       with_string_0(source, GLO(pathname_encoding), source_asciz, {
         with_string_0(dest, GLO(pathname_encoding), dest_asciz,
                       { hardlink_file(source_asciz,dest_asciz); });
@@ -1291,4 +1607,80 @@ DEFUN(POSIX::SHORTCUT-INFO, file)
  fail_psl: psl->lpVtbl->Release(psl);
  fail_none: end_system_call(); OS_file_error(*file);
 }
+
+DEFUN(POSIX::SYSTEM-INFO,)
+{ /* interface to GetSystemInfo() */
+  SYSTEM_INFO si;
+  begin_system_call();
+  GetSystemInfo(&si);
+  end_system_call();
+  switch (si.wProcessorArchitecture) {
+    case PROCESSOR_ARCHITECTURE_UNKNOWN: pushSTACK(`:UNKNOWN`); break;
+    case PROCESSOR_ARCHITECTURE_INTEL:   pushSTACK(`:INTEL`); break;
+    case PROCESSOR_ARCHITECTURE_MIPS:    pushSTACK(`:MIPS`); break;
+    case PROCESSOR_ARCHITECTURE_ALPHA:   pushSTACK(`:ALPHA`); break;
+    case PROCESSOR_ARCHITECTURE_PPC:     pushSTACK(`:PPC`); break;
+    case PROCESSOR_ARCHITECTURE_IA64 :   pushSTACK(`:IA64`); break;
+    default: pushSTACK(UL_to_I(si.wProcessorArchitecture));
+  }
+  pushSTACK(UL_to_I(si.dwPageSize));
+  pushSTACK(UL_to_I((DWORD)si.lpMinimumApplicationAddress));
+  pushSTACK(UL_to_I((DWORD)si.lpMaximumApplicationAddress));
+  pushSTACK(UL_to_I(si.dwActiveProcessorMask));
+  pushSTACK(UL_to_I(si.dwNumberOfProcessors));
+  pushSTACK(UL_to_I(si.dwAllocationGranularity));
+  pushSTACK(fixnum(si.wProcessorLevel));
+  pushSTACK(fixnum(si.wProcessorRevision));
+  funcall(`POSIX::MAKE-SYSTEM-INFO`,9);
+}
+
+DEFUN(POSIX::VERSION,)
+{ /* interface to GetVersionEx() */
+  OSVERSIONINFOEX vi;
+  vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
+  begin_system_call();
+  if (!GetVersionEx((OSVERSIONINFO*)&vi)) OS_error();
+  end_system_call();
+
+  pushSTACK(UL_to_I(vi.dwMajorVersion));
+  pushSTACK(UL_to_I(vi.dwMinorVersion));
+  pushSTACK(UL_to_I(vi.dwBuildNumber));
+  switch (vi.dwPlatformId) {
+    case VER_PLATFORM_WIN32s:        pushSTACK(`:S`); break;
+    case VER_PLATFORM_WIN32_WINDOWS: pushSTACK(`:WINDOWS`); break;
+    case VER_PLATFORM_WIN32_NT:      pushSTACK(`:NT`); break;
+    default: pushSTACK(UL_to_I(vi.dwPlatformId));
+  }
+  pushSTACK(asciz_to_string(vi.szCSDVersion,GLO(misc_encoding)));
+  pushSTACK(UL_to_I(vi.wServicePackMajor));
+  pushSTACK(UL_to_I(vi.wServicePackMinor));
+  { /* wSuiteMask */
+    object suites = NIL;
+    unsigned int count = 0;
+    if (vi.wSuiteMask & VER_SUITE_BACKOFFICE)
+      { pushSTACK(`:BACKOFFICE`); count++; }
+    if (vi.wSuiteMask & VER_SUITE_DATACENTER)
+      { pushSTACK(`:DATACENTER`); count++; }
+    if (vi.wSuiteMask & VER_SUITE_ENTERPRISE)
+      { pushSTACK(`:ENTERPRISE`); count++; }
+    if (vi.wSuiteMask & VER_SUITE_SMALLBUSINESS)
+      { pushSTACK(`:SMALLBUSINESS`); count++; }
+    if (vi.wSuiteMask & VER_SUITE_SMALLBUSINESS_RESTRICTED)
+      { pushSTACK(`:SMALLBUSINESS-RESTRICTED`); count++; }
+    if (vi.wSuiteMask & VER_SUITE_TERMINAL)
+      { pushSTACK(`:TERMINAL`); count++; }
+    if (vi.wSuiteMask & VER_SUITE_PERSONAL)
+      { pushSTACK(`:PERSONAL`); count++; }
+    if (count) suites = listof(count);
+    pushSTACK(suites);
+  }
+  switch (vi.wProductType) {
+    case VER_NT_WORKSTATION:       pushSTACK(`:WORKSTATION`); break;
+    case VER_NT_DOMAIN_CONTROLLER: pushSTACK(`:DOMAIN-CONTROLLER`); break;
+    case VER_NT_SERVER:            pushSTACK(`:SERVER`); break;
+    default: pushSTACK(UL_to_I(vi.wProductType));
+  }
+  funcall(`POSIX::MAKE-VERSION`,9);
+}
+
 #endif
