@@ -1,9 +1,10 @@
 ;; Module for wildcard matching in CLISP
 ;; Bruno Haible 18.4.1995
+;; Sam Steingold 2001-2004
 
 (defpackage "WILDCARD"
   (:use "FFI" "COMMON-LISP")
-  (:export "MATCH"))
+  (:export "MATCH" "WILDCARD-MATCHER"))
 (in-package "WILDCARD")
 
 (default-foreign-language :stdc)
@@ -35,3 +36,7 @@
   (zerop
     (fnmatch pattern string
              (logior FNM_PATHNAME (if case-insensitive FNM_CASEFOLD 0)))))
+
+(defun wildcard-matcher (pattern)
+  "A valid value for *APROPOS-MATCHER*."
+  (lambda (name) (zerop (fnmatch pattern name FNM_CASEFOLD))))
