@@ -18,8 +18,6 @@
 
 ;; Das temporäre File, das LISP beim Editieren anlegt:
 (defun editor-tempfile ()
-  #+OS/2 "lisptemp.lisp"
-  #+AMIGA "T:lisptemp.lisp"
   #+(or UNIX WIN32) (merge-pathnames "lisptemp.lisp" (user-homedir-pathname))
 )
 
@@ -31,34 +29,16 @@
   ) )
   ; Damit TRUENAME keinen Fehler liefert, wenn das File noch nicht existiert,
   ; stellen wir sicher, dass das File existiert:
-  #+(or UNIX AMIGA ACORN-RISCOS)
+  #+UNIX
   (unless (probe-file file)
     (close (open file :direction :output))
   )
-  #+(or OS/2 WIN32)
+  #+WIN32
     (execute (editor-name) ; das ist der Name des Editors
              (namestring file t) ; file als String
     )
   #+UNIX
     (shell (format nil "~A ~A" (editor-name) (truename file)))
-  #+AMIGA
-    (shell (format nil "~A \"~A\"" (editor-name) (truename file)))
-  #+ACORN-RISCOS
-    (let ((pathname (truename file)))
-      (shell
-        (format nil "~A ~A"
-                    (editor-name)
-                    (if (pathname-type pathname)
-                      ; swap pathname's name and type
-                      (merge-pathnames
-                        (make-pathname :name (pathname-type pathname)
-                                       :type (pathname-name pathname)
-                        )
-                        pathname
-                      )
-                      pathname
-                    )
-    ) ) )
 )
 
 (defun ed (&optional arg &aux funname sym fun def)
